"""Copyright 2021 Google LLC.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.

You may obtain a copy of the License at
                https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import copy
import json
import logging
import time
import unittest
from functools import wraps
from typing import Any, Dict, List, Optional, Tuple, Union

from pyangbind.lib import pybindJSON
from pyangbind.lib.base import PybindBase
from retry.api import retry_call

from oc_config_validate import context, schema, target
from oc_config_validate.gnmi import gnmi_pb2  # type: ignore

# Switch to enable logging of gNMI messages in the testcase logs.
LOG_GNMI = False


def failfast(method):
    """Wrap a test to stop upon failure."""
    @wraps(method)
    def inner(self, *args, **kw):
        if getattr(self, 'failfast', False):
            self.stop()
        method(self, *args, **kw)
    return inner


def retryAssertionError(method):
    """Wrap a test to retry if AssertionError."""
    @wraps(method)
    def inner(self, *args, **kw):
        if hasattr(self, 'retries'):
            delay = getattr(self, 'retry_delay', 10)
            retry_call(method,  fargs=[self, *args], fkwargs=kw,
                       tries=getattr(self, 'retries'), delay=delay)
        else:
            method(self, *args, **kw)
    return inner


class TestSuite(unittest.suite.TestSuite):
    """TestSuite from which all other suites inherit.

    Allows inserting arguments to the Tests, which are made attributes of the
        class.
    """

    # Instructs parent class to not cleanup after performing tests.
    _cleanup = False
    test_name = ""
    test_class_name = ""

    def insertArgs(self, args: Dict[str, Any]):
        """Insert test arguments to all test cases in this suite.

        The keys of the argument dictionary are made attributes of the class.

        Args:
            args: dictionary of arguments.
        """
        for test in self._tests:
            TestCase.insertArgs(test, args)

    def insertTarget(self, tgt: target.TestTarget):
        """Insert the gNMI Target to all test cases in this suite.

        Args:
            tgt: A gNMI Target.
        """
        for test in self._tests:
            TestCase.insertTarget(test, tgt)


class TestCase(unittest.case.TestCase):
    """TestCase from which all other tests inherit.

    self.test_target is the gNMI Target to test.

    """
    result = None
    test_target = None
    duration_sec = 0
    start_time_sec = 0

    def phony(self):
        """A phony test method to be used for testing."""

    def run(self, result):
        """Hook the result such that it can be measured and logged."""
        logging.debug('Starting test %s ...', self)
        self.result = result
        if result:
            self.start_time_sec = int(time.time())
            super().run(result)
            self.duration_sec = int(time.time() - self.start_time_sec)
            logging.debug('Finished %s in %d secs', self, self.duration_sec)

    def log(self, formatter, *args):
        """Log a message during a test.

        Args:
            formatter: The base python string formatter.
            *args: The formatter parameters. Allows logging messages during
                test executions.
        """
        if self.result:
            self.result.log(formatter, *args)
        logging.debug(formatter, *args)

    def gNMIGet(self, xpath: str) -> Optional[gnmi_pb2.TypedValue]:
        """Send a gNMI Get message to the test's target.

        If the path produces several values in the response, only the first is
          returned.

        Args:
            xpath:  gNMI Path.

        Returns:
            A single response value, or None if error.
        """
        try:
            resp = self.test_target.gNMIGet(xpath)
        except Exception as err:
            self.log("Get(%s) <= Error: %s", xpath, err)
            return None
        try:
            resp_val = resp.notification[0].update[0].val
        except IndexError:
            self.log("Get(%s) <= Bad response: %s", xpath, resp)
            return None
        if LOG_GNMI:
            msg = ("gNMI Get(%s) <= %s", xpath, resp_val)
            self.log(*msg)
            logging.info(*msg)
        return resp_val

    def gNMIGetAssertJson(self, xpath: str) -> Union[str, bytes]:
        """Send a gNMI Get and asserts that the response is a JSON text.

        Args:
            xpath:  gNMI Path.

        Returns:
            a JSON text.

        Raises:
          AssertionError if unable to gNMI Get or if the response is not JSON.
        """
        resp = self.gNMIGetAssertGet(xpath)
        resp_val = resp.json_ietf_val
        self.assertIsNotNone(resp_val,
                             "The gNMI GET response is not JSON IETF")
        return resp_val

    def gNMIGetAssertBool(self, xpath: str) -> bool:
        """Send a gNMI Get and asserts that the response is a bool value.

        Args:
            xpath:  gNMI Path.

        Returns:
            a boolean value.

        Raises:
          AssertionError if unable to gNMI Get or if the response is not bool.
        """
        resp = self.gNMIGetAssertGet(xpath)
        self.assertIsNotNone(
            resp.bool_val,
            "The gNMI Get response is not a boolean: %s" % resp)
        return resp

    def gNMIGetAssertGet(self, xpath: str) -> gnmi_pb2.TypedValue:
        """Send a gNMI Get and asserts that there is a response.

        Args:
            xpath:  gNMI Path.

        Returns:
            the response to the Get requets.

        Raises:
          AssertionError if unable to gNMI Get or if the response is None.
        """
        resp = self.gNMIGet(xpath)
        self.assertIsNotNone(resp, "No gNMI GET response")
        return resp

    def gNMISetUpdate(self, xpath: str, value: Any) -> bool:
        """Send a gNMI Set Update message to the test's target

        Args:
            xpath: The gNMI path to update.
            value: Value to set; can be numeric, string or JSON-IETF.

        Returns:
          False if the gNMI Set did not succeed.
        """
        if LOG_GNMI:
            msg = ("gNMI Set Update(%s) => %s", xpath, value)
            self.log(*msg)
            logging.info(*msg)
        try:
            self.test_target.gNMISetUpdate(xpath, value)
        except Exception as err:
            self.log("Set(%s) <= Error: %s", xpath, err)
            return False
        return True

    def gNMISetDelete(self, xpath: str) -> bool:
        """Send a gNMI Set Delete message to the test's target

        Args:
            xpath: The gNMI path to delete.

        Args:
            xpath:  gNMI Path.

        Returns:
            A single response value, or None if error.
        """
        if LOG_GNMI:
            msg = ("gNMI Set Delete(%s) =>", xpath)
            self.log(*msg)
            logging.info(*msg)
        try:
            self.test_target.gNMISetDelete(xpath)
        except Exception as err:
            self.log("Set(%s) <= Error: %s", xpath, err)
            return False
        return True

    def gNMISubsOnce(self, xpaths: List[str]) -> Optional[
            List[gnmi_pb2.Notification]]:
        """Send a gNMI Subscribe message using ONCE mode.

        Gets all the Notification messages that came as response.

        Args:
            xpaths: List of gNMI paths to subscribe to.

        Returns:
          A list of Notifications received, or None if error.
        """
        try:
            resp = self.test_target.gNMISubsOnce(xpaths)
        except Exception as err:
            self.log("SubscribeOnce(%s) <= Error: %s", xpaths, err)
            return None
        if LOG_GNMI:
            msg = ("gNMI SubscribeOnce(%s) <= %s", xpaths,
                   schema.notificationsJsonString(resp))
            self.log(*msg)
            logging.info(*msg)
        return resp

    def gNMISubsStreamSample(
            self, xpath: str, sample_interval: int, timeout: int) -> Optional[
            List[gnmi_pb2.Notification]]:
        """Send a gNMI Subscribe message, using STREAM SAMPLE mode.

        Gets all the Notification messages that came as response. It returns
            a dictionary of all the Updates, keyed by Update path.

        Args:
            xpath: gNMI path to subscribe to.
            sample_interval: Nanoseconds between updates.
            timeout: Seconds to keep the gRPC channel open and receive
                    updates.

        Returns:
          A list of Notifications received, or None if error.
        """
        try:
            resp = self.test_target.gNMISubsStreamSample(
                xpath, sample_interval, timeout)
        except Exception as err:
            self.log("SubscribeStream(%s) <= Error: %s", xpath, err)
            return None
        if LOG_GNMI:
            msg = ("gNMI SubscribeSample(%s) <= %s", xpath,
                   schema.notificationsJsonString(resp))
            self.log(*msg)
            logging.info(*msg)
        return resp

    def gNMISubsStreamOnChange(
            self,
            xpath: str,
            timeout: int,
            sync_response_callback=None) -> Tuple[
                Optional[List[gnmi_pb2.Notification]],
                Optional[List[gnmi_pb2.Notification]]]:
        """Subscribes on-change and returns the Notifications received.

        It returns the Updates before and after receiving a sync_response.

        All Updates received are accumulated up to when the channel is
         closed (by server, by timeout or error).

        Optionally, it calls a callback function when sync_response is
          received.

        Does not support heartbeat. nor OnlyUpdates (the initial update
          is needed for value change comparison).

        Args:
            xpath: gNMI path to subscribe to.
            timeout: Seconds to keep the gRPC channel open.
            sync_response_callback: Function to call upon sync_response.

        Returns:
            2 Lists of gnmi_pb2.Notification objects received,
              before and after sync_response.
        """

        try:
            before, after = self.test_target.gNMISubsStreamOnChange(
                xpath, timeout, sync_response_callback)
        except Exception as err:
            self.log("SubscribeOnChange(%s) <= Error: %s", xpath, err)
            return None, None
        if LOG_GNMI:
            msg = ("gNMI SubscribeOnChange(%s) <= %s\n", xpath,
                   schema.notificationsJsonString(before + after))
            self.log(*msg)
            logging.info(*msg)
        return before, after

    @classmethod
    def insertArgs(cls, test: unittest.TestCase, args: Dict[str, Any]):
        """Insert test arguments to the test.

        The keys of the argument dictionary are made attributes of the class.

        Args:
            args: dictionary of strings keyed by argument name.
        """
        for key, value in args.items():
            setattr(test, key, value)

    @classmethod
    def insertTarget(cls,
                     test: unittest.TestCase,
                     tgt: target.TestTarget):
        """Insert the gNMI Target to the test.

        Args:
            tgt: A gNMI Target.
        """
        setattr(test, "test_target", tgt)

    def assertArgs(self,
                   args: List[str]):
        """Assert that the arguments are present in the class.

        Args:
            args: A list of argument names to check.

        Raises:
            AssertionError if any argument is not in the class.
        """
        for arg in args:
            self.assertIsNotNone(getattr(self, arg, None),
                                 "Missing class argument: %s" % arg)

    def assertJsonModel(self, json_value: Union[str, bytes],
                        model: PybindBase, msg: str):
        """Assert that the JSON text complies to the model schema.

        Args:
            json_value: JSON-IETF to check.
            model: the OC model class name in the oc_config_validate.models
              package, as `module.class`.
            xpath: the xpath to the OC container to create.
            mgs: Message to show when the check fails.

        Raises:
          AssertionError if the JSON text does not match the model.
        """
        match, error = True, None
        model_obj = copy.copy(model)
        try:
            schema.decodeJson(json_value, model_obj)
        except (AttributeError, ValueError) as err:
            match, error = False, err
        self.assertTrue(match, "%s: %s" % (msg, error))

    def assertJsonCmp(self, json_value: Union[str, bytes], want: dict):
        """Assert that the JSON response matches the wanted values..

        Args:
            json_value: JSON-IETF gNMI response to check.
            want: Dictionary of wanted values.

        Raises:
          AssertionError if unable to compare, of the the wanted values are not
            present in the response.
        """
        got = json.loads(json_value)
        cmp, diff = schema.intersectCmp(
            json.loads(pybindJSON.dumps(want, mode='ietf')), got)
        self.assertTrue(cmp, diff)

    def assertXpath(self, xpath: str):
        """Assert that the xpath is valid.

        Args:
            xpath:  gNMI Path.

        Raises:
          AssertionError if xpath is not valid
        """
        self.assertTrue(schema.isPathOK(xpath), "'%s' not a gNMI path" % xpath)

    def assertModelXpath(self, model: str, xpath: str):
        """Assert that the model and xpath correspond to a valid OC container.

        Args:
            model: the OC model class name in the oc_config_validate.models
              package, as `module.class`.
            xpath: the xpath to the OC container to create.

        Raises: AssertionError.
        """
        match, error = True, None
        try:
            schema.ocContainerFromPath(model, xpath)
        except (AttributeError, schema.BaseError) as err:
            match, error = False, err
        self.assertTrue(
            match,
            "xpath %s does not point to a container in model %s: %s" % (
                xpath, model, error))


class TestResult(unittest.TestResult):
    """TestResult allows logging messages during the tests."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.successes = []
        self.log_message = ''
        self.test_name = 'Test'
        self.duration_sec = 0

    def log(self, formatter, *args):
        """Log a message during a test.

        Args:
            formatter: The base python string formatter.
            *args: The formatter parameters. Allows logging messages during
                test executions.
        """
        self.log_message += formatter % args + '\n'

    def startTest(self, test: TestCase):  # pytype: disable=signature-mismatch
        """Clean the log before every test."""
        super().startTest(test)
        self.log_message = ''

    @failfast
    def addError(self, test: TestCase, err):
        """Override parent addError to support logging."""
        log = '%s\n%s' % (
            self.log_message, self._exc_info_to_string(err, test))  # pytype: disable=attribute-error # noqa
        self.errors.append((test, log))
        self._mirrorOutput = True
        logging.error("%s - ERR:\n%s", test, log)

    @failfast
    def addFailure(self, test: TestCase, err):
        """Override parent addFailure to support logging."""
        log = '%s\n %s' % (
            self.log_message, self._exc_info_to_string(err, test))  # pytype: disable=attribute-error # noqa
        self.failures.append((test, log))
        self._mirrorOutput = True
        logging.error("%s - FAIL:\n%s", test, log)

    def addSuccess(self, test: TestCase):   # pytype: disable=signature-mismatch  # noqa
        """Store the log messages for every successful test."""
        self.successes.append((test, self.log_message))
        logging.debug("%s - PASS:\n%s", test, self.log_message)

    def addSkip(self, test: TestCase, reason: str):   # pytype: disable=signature-mismatch  # noqa
        """Override parent addFailure to support logging."""
        self.skipped.append((test, reason))
        logging.debug("%s - SKIP:\n%s", test, reason)


class MethodResult():
    """Result of a method called in a TestCase"""

    def __init__(self, test_method: Tuple[TestCase, str], result: str):
        """Create a Result.

        Args:
          test_method: A tuple of TestCase and logs.
          result: A text for the result, as PASS or FAIL.
        """
        test_case, log = test_method
        split_trace = test_case.id().split('.')
        self.test_case = split_trace[-1]
        self.test_class = split_trace[-2]
        self.test_module = '.'.join(split_trace[0:-3])
        self.start_time_sec = test_case.start_time_sec
        self.duration_sec = test_case.duration_sec
        self.result = result
        self.log = log


class TestcaseResult():
    """Result of a TestCase"""

    def __init__(self, test_result):
        """Create a Result.

        Args:
          test_result: A TestResult object.

        """
        self.test_name = test_result.test_name
        self.success = test_result.wasSuccessful()
        self.duration_sec = test_result.duration_sec
        self.results = []
        for m in test_result.successes:
            self.results.append(MethodResult(m, 'PASS'))
        for m in test_result.failures:
            self.results.append(MethodResult(m, 'FAIL'))
        for m in test_result.errors:
            self.results.append(MethodResult(m, 'ERROR'))
        for m in test_result.skipped:
            self.results.append(MethodResult(m, 'SKIPPED'))


class TestRun():
    """Results of a completed test run.

    Including time, context and method details.

    Args:
        target: gNMI Target against which the tests run.
        ctx: Test Context.

    """
    description = ""
    labels = []
    results = []
    start_time_sec = 0
    end_time_sec = 0
    tests_pass = 0
    tests_total = 0
    tests_fail = 0

    def __init__(self, ctx: context.TestContext, tgt: target.TestTarget):
        self.test_target = ctx.target.target
        if ctx.description:
            self.description = ctx.description
        if ctx.labels:
            self.labels = ctx.labels.copy()

    def copyResults(self, results: List[TestResult], start_time_sec: float,
                    end_time_sec: float):
        """Copy the run test results.

        Args:
            results: A list of test results to copy.
            start_time_sec: EPOCH time when the tests started.
            end_time_sec: EPOCH time when the tests ended.
        """

        self.start_time_sec = start_time_sec
        self.end_time_sec = end_time_sec
        self.results = []
        n_pass = 0
        for result in results:
            n_pass += result.wasSuccessful()
            self.results.append(TestcaseResult(result))
        self.tests_pass = n_pass
        self.tests_total = len(results)
        self.tests_fail = self.tests_total - self.tests_pass

    def summary(self) -> str:
        """Return a text with the results summary. 'PASS x, FAIL y'.
        """

        return "PASS %d, FAIL %d" % (self.tests_pass, self.tests_fail)



"""Test cases for subinterfaces with IPv4 addresses."""

import json

from pyangbind.lib import pybindJSON
from retry import retry

from oc_config_validate import schema, testbase
from oc_config_validate.models.interfaces import \
    openconfig_interfaces as oc_interfaces


class SetSubifDhcp(testbase.TestCase):
    """Tests configuring DHCP on a subinterface.

    1. A gNMI Set message is sent to configure the subinterface.
    1. A gNMI Get message on the /config container validates it.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        dhcp: True to enable DHCP, defaults to False.
    """
    interface = ""
    index = 0
    dhcp = False

    def test0100(self):
        self.assertArgs(["interface"])

        iface = oc_interfaces().interfaces.interface.add(self.interface)
        subif = iface.subinterfaces.subinterface.add(self.index)
        if self.index:
            subif.vlan.config.vlan_id = self.index
        subif.ipv4.config.dhcp_client = self.dhcp
        xpath = "/interfaces/interface[name=%s]" % self.interface
        _json_value = json.loads(pybindJSON.dumps(iface, mode='ietf'))
        schema.fixSubifIndex(_json_value)
        self.assertTrue(self.gNMISetUpdate(xpath, _json_value),
                        "gNMI Set did not succeed.")

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0200(self):
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/config/dhcp-client") % (
            self.interface, self.index)
        resp_val = self.gNMIGetAssertBool(xpath)
        self.assertEqual(resp_val, self.dhcp,
                         "Dhcp client config = %s, wanted %s" % (
                             resp_val, self.dhcp))

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0300(self):
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/state/dhcp-client") % (
            self.interface, self.index)
        resp_val = self.gNMIGetAssertBool(xpath)
        self.assertEqual(resp_val, self.dhcp,
                         "Dhcp client state = %s, wanted %s" % (
                             resp_val, self.dhcp))


class CheckSubifDhcpState(testbase.TestCase):
    """Checks the DHCP state on a subinterface.

    1. A gNMI Get message on the /state container.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        dhcp: True to enable DHCP, defaults to False.
    """
    interface = ""
    index = 0
    dhcp = False

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0100(self):
        """"""
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/state/dhcp-client") % (
            self.interface, self.index)
        resp_val = self.gNMIGetAssertBool(xpath)
        self.assertEqual(resp_val, self.dhcp,
                         "Dhcp client state = %s, wanted %s" % (
                             resp_val, self.dhcp))


class CheckSubifDhcpConfig(testbase.TestCase):

    """Checks the DHCP config on a subinterface.

    1. A gNMI Get message on the /config container.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        dhcp: True to enable DHCP, defaults to False.
    """
    interface = ""
    index = 0
    dhcp = False

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0100(self):
        """"""
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/config/dhcp-client") % (
            self.interface, self.index)
        resp_val = self.gNMIGetAssertBool(xpath)
        self.assertEqual(resp_val, self.dhcp,
                         "Dhcp client config = %s, wanted %s" % (
                             resp_val, self.dhcp))


class AddSubifIp(testbase.TestCase):
    """Tests configuring an IP on a subinterface.

    1. A gNMI Set message is sent to configure the subinterface.
    1. A gNMI Get message on the /config container validates it.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        address: IPv4 address to add.
        prefix_length: Prefix lenght of the IPv4 address to add.
    """
    interface = ""
    index = 0
    address = ""
    prefix_length = 32
    iface = None

    def setUp(self):
        self.iface = oc_interfaces().interfaces.interface.add(self.interface)
        subif = self.iface.subinterfaces.subinterface.add(self.index)
        if self.index:
            subif.vlan.config.vlan_id = self.index
        self.addr = subif.ipv4.addresses.address.add(self.address)
        self.addr.config.ip = self.address
        self.addr.config.prefix_length = self.prefix_length

    def test0100(self):
        self.assertArgs(["interface", "address", "prefix_length"])
        xpath = "/interfaces/interface[name=%s]" % self.interface
        _json_value = json.loads(pybindJSON.dumps(self.iface, mode='ietf'))
        schema.fixSubifIndex(_json_value)
        self.assertTrue(self.gNMISetUpdate(xpath, _json_value),
                        "gNMI Set did not succeed.")

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0200(self):
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/addresses"
                 "/address[ip=%s]/config") % (
            self.interface, self.index, self.address)
        resp_val = self.gNMIGetAssertJson(xpath)
        self.assertJsonModel(
            resp_val, self.addr.config,
            'gNMI Get on the /config container does not match model')
        want = json.loads(
            schema.removeOpenConfigPrefix(
                pybindJSON.dumps(self.addr.config, mode='ietf')))
        self.assertJsonCmp(resp_val, want)

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0300(self):
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/addresses"
                 "/address[ip=%s]/state") % (
            self.interface, self.index, self.address)
        resp_val = self.gNMIGetAssertJson(xpath)

        self.addr.state._set_ip(self.address)
        self.addr.state._set_prefix_length(self.prefix_length)
        self.assertJsonModel(
            resp_val, self.addr.state,
            'gNMI Get on the /state container does not match model')
        want = json.loads(
            schema.removeOpenConfigPrefix(
                pybindJSON.dumps(self.addr.state, mode='ietf')))
        self.assertJsonCmp(resp_val, want)


class RemoveSubifIp(testbase.TestCase):
    """Tests removing an IP on a subinterface.

    1. gNMI Get message on the /config container, to check it is configured.
    1. gNMI Set message to delete the ip.
    1. gNMI Get message on the /config container to check it is not there.
    1. gNMI Get message on the /state container to check it is not there.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        address: IPv4 address to remove.
        prefix_length: Prefix lenght of the IPv4 address to remove.
    """
    interface = ""
    index = 0
    address = ""
    prefix_length = 32
    xpath = ""

    def test0100(self):
        self.assertArgs(["interface", "address", "prefix_length"])
        self.xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                      "subinterface[index=%d]/ipv4/addresses"
                      "/address[ip=%s]") % (
            self.interface, self.index, self.address)
        if not self.gNMIGet(self.xpath + "/config"):
            self.log("IP %s not configured at %s.%s",
                     self.address, self.interface, self.index)
            return
        self.assertTrue(self.gNMISetDelete(self.xpath),
                        "gNMI Delete did not succeed.")

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0200(self):
        resp = self.gNMIGet(self.xpath + "/config")
        self.assertIsNone(
            resp,
            "IP %s still configured at %s.%s" % (self.address,
                                                 self.interface, self.index))

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0300(self):
        resp = self.gNMIGet(self.xpath + "/state")
        self.assertIsNone(
            resp,
            "IP %s still at %s.%s" % (self.address,
                                      self.interface, self.index))


class CheckSubifIpState(testbase.TestCase):
    """Checks the state on an ip address configured on a subinterface.

    1. A gNMI Get message on the /state container.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        address: IPv4 address.
        prefix_length: Prefix lenght of the IPv4 address.
    """
    interface = ""
    index = 0
    address = ""
    prefix_length = 32

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0100(self):
        """"""
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/addresses/"
                 "address[ip=%s]/state") % (
            self.interface, self.index, self.address)
        resp_val = self.gNMIGetAssertJson(xpath)
        iface = oc_interfaces().interfaces.interface.add(self.interface)
        subif = iface.subinterfaces.subinterface.add(self.index)
        if self.index:
            subif.vlan.config.vlan_id = self.index
        addr = subif.ipv4.addresses.address.add(self.address)
        addr.state._set_ip(self.address)
        addr.state._set_prefix_length(self.prefix_length)
        self.assertJsonModel(
            resp_val, addr.state,
            'gNMI Get on the /state container does not match model')
        want = json.loads(
            schema.removeOpenConfigPrefix(
                pybindJSON.dumps(addr.state, mode='ietf')))
        self.assertJsonCmp(resp_val, want)


class CheckSubifIpConfig(testbase.TestCase):
    """Checks the config on an ip address on a subinterface.

    1. A gNMI Get message on the /config container.

    All arguments are read from the Test YAML description.

    Args:
        interface: Name of the physical interface.
        index: Index of the subinterface, defaults to 0.
        address: IPv4 address.
        prefix_length: Prefix lenght of the IPv4 address.
    """
    interface = ""
    index = 0
    address = ""
    prefix_length = 32

    @retry(exceptions=AssertionError, tries=5, delay=15)
    def test0100(self):
        """"""
        xpath = ("/interfaces/interface[name=%s]/subinterfaces/"
                 "subinterface[index=%d]/ipv4/addresses/"
                 "address[ip=%s]/config") % (
            self.interface, self.index, self.address)
        resp_val = self.gNMIGetAssertJson(xpath)
        iface = oc_interfaces().interfaces.interface.add(self.interface)
        subif = iface.subinterfaces.subinterface.add(self.index)
        if self.index:
            subif.vlan.config.vlan_id = self.index
        addr = subif.ipv4.addresses.address.add(self.address)
        addr.config.ip = self.address
        addr.config.prefix_length = self.prefix_length
        self.assertJsonModel(
            resp_val, addr.config,
            'gNMI Get on the /config container does not match model')
        want = json.loads(
            schema.removeOpenConfigPrefix(
                pybindJSON.dumps(addr.config, mode='ietf')))
        self.assertJsonCmp(resp_val, want)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gnmi_ext.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0egnmi_ext.proto\x12\x08gnmi_ext\"\xac\x01\n\tExtension\x12\x37\n\x0eregistered_ext\x18\x01 \x01(\x0b\x32\x1d.gnmi_ext.RegisteredExtensionH\x00\x12\x39\n\x12master_arbitration\x18\x02 \x01(\x0b\x32\x1b.gnmi_ext.MasterArbitrationH\x00\x12$\n\x07history\x18\x03 \x01(\x0b\x32\x11.gnmi_ext.HistoryH\x00\x42\x05\n\x03\x65xt\"E\n\x13RegisteredExtension\x12!\n\x02id\x18\x01 \x01(\x0e\x32\x15.gnmi_ext.ExtensionID\x12\x0b\n\x03msg\x18\x02 \x01(\x0c\"Y\n\x11MasterArbitration\x12\x1c\n\x04role\x18\x01 \x01(\x0b\x32\x0e.gnmi_ext.Role\x12&\n\x0b\x65lection_id\x18\x02 \x01(\x0b\x32\x11.gnmi_ext.Uint128\"$\n\x07Uint128\x12\x0c\n\x04high\x18\x01 \x01(\x04\x12\x0b\n\x03low\x18\x02 \x01(\x04\"\x12\n\x04Role\x12\n\n\x02id\x18\x01 \x01(\t\"S\n\x07History\x12\x17\n\rsnapshot_time\x18\x01 \x01(\x03H\x00\x12$\n\x05range\x18\x02 \x01(\x0b\x32\x13.gnmi_ext.TimeRangeH\x00\x42\t\n\x07request\"\'\n\tTimeRange\x12\r\n\x05start\x18\x01 \x01(\x03\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x03*3\n\x0b\x45xtensionID\x12\r\n\tEID_UNSET\x10\x00\x12\x15\n\x10\x45ID_EXPERIMENTAL\x10\xe7\x07\x42+Z)github.com/openconfig/gnmi/proto/gnmi_extb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'gnmi_ext_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z)github.com/openconfig/gnmi/proto/gnmi_ext'
  _EXTENSIONID._serialized_start=549
  _EXTENSIONID._serialized_end=600
  _EXTENSION._serialized_start=29
  _EXTENSION._serialized_end=201
  _REGISTEREDEXTENSION._serialized_start=203
  _REGISTEREDEXTENSION._serialized_end=272
  _MASTERARBITRATION._serialized_start=274
  _MASTERARBITRATION._serialized_end=363
  _UINT128._serialized_start=365
  _UINT128._serialized_end=401
  _ROLE._serialized_start=403
  _ROLE._serialized_end=421
  _HISTORY._serialized_start=423
  _HISTORY._serialized_end=506
  _TIMERANGE._serialized_start=508
  _TIMERANGE._serialized_end=547
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from operator import attrgetter
from pyangbind.lib.yangtypes import RestrictedPrecisionDecimalType
from pyangbind.lib.yangtypes import RestrictedClassType
from pyangbind.lib.yangtypes import TypedListType
from pyangbind.lib.yangtypes import YANGBool
from pyangbind.lib.yangtypes import YANGListType
from pyangbind.lib.yangtypes import YANGDynClass
from pyangbind.lib.yangtypes import ReferenceType
from pyangbind.lib.base import PybindBase
from collections import OrderedDict
from decimal import Decimal
from bitarray import bitarray
import six

# PY3 support of some PY2 keywords (needs improved)
if six.PY3:
  import builtins as __builtin__
  long = int
elif six.PY2:
  import __builtin__

class openconfig_rib_bgp_ext(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-ext - based on the path /openconfig-rib-bgp-ext. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Defines additional data nodes for the OpenConfig BGP RIB model.
These items reflect extensions that are desirable features but
are not currently supported in a majority of BGP
implementations.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_tables(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-tables - based on the path /openconfig-rib-bgp-tables. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains structural data definitions for
BGP routing tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-attributes - based on the path /openconfig-rib-bgp-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for BGP
attributes for use in BGP RIB tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_shared_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-shared-attributes - based on the path /openconfig-rib-bgp-shared-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains structural data definitions for
attribute sets shared across routes.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_table_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-table-attributes - based on the path /openconfig-rib-bgp-table-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for data
related to a RIB entry, or RIB table.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp - based on the path /openconfig-rib-bgp. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Defines a data model for representing BGP routing table (RIB)
contents.  The model supports 5 logical RIBs per address family:

loc-rib: This is the main BGP routing table for the local routing
instance, containing best-path selections for each prefix. The
loc-rib table may contain multiple routes for a given prefix,
with an attribute to indicate which was selected as the best
path.  Note that multiple paths may be used or advertised even if
only one path is marked as best, e.g., when using BGP
add-paths.  An implementation may choose to mark multiple
paths in the RIB as best path by setting the flag to true for
multiple entries.

adj-rib-in-pre: This is a per-neighbor table containing the NLRI
updates received from the neighbor before any local input policy
rules or filters have been applied.  This can be considered the
'raw' updates from a given neighbor.

adj-rib-in-post: This is a per-neighbor table containing the
routes received from the neighbor that are eligible for
best-path selection after local input policy rules have been
applied.

adj-rib-out-pre: This is a per-neighbor table containing routes
eligible for sending (advertising) to the neighbor before output
policy rules have been applied.

adj-rib-out-post: This is a per-neighbor table containing routes
eligible for sending (advertising) to the neighbor after output
policy rules have been applied.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_tables(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-tables - based on the path /openconfig-rib-bgp-tables. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains structural data definitions for
BGP routing tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_shared_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-shared-attributes - based on the path /openconfig-rib-bgp-shared-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains structural data definitions for
attribute sets shared across routes.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-attributes - based on the path /openconfig-rib-bgp-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for BGP
attributes for use in BGP RIB tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_table_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-table-attributes - based on the path /openconfig-rib-bgp-table-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for data
related to a RIB entry, or RIB table.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-attributes - based on the path /openconfig-rib-bgp-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for BGP
attributes for use in BGP RIB tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_shared_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-shared-attributes - based on the path /openconfig-rib-bgp-shared-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains structural data definitions for
attribute sets shared across routes.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-attributes - based on the path /openconfig-rib-bgp-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for BGP
attributes for use in BGP RIB tables.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_table_attributes(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-table-attributes - based on the path /openconfig-rib-bgp-table-attributes. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: This submodule contains common data definitions for data
related to a RIB entry, or RIB table.
  """
  _pyangbind_elements = {}

  

class openconfig_rib_bgp_types(PybindBase):
  """
  This class was auto-generated by the PythonClass plugin for PYANG
  from YANG module openconfig-rib-bgp-types - based on the path /openconfig-rib-bgp-types. Each member element of
  the container is represented as a class variable - with a specific
  YANG type.

  YANG Description: Defines identity and type defintions associated with
the OpenConfig BGP RIB modules
  """
  _pyangbind_elements = {}

  




from oc_config_validate.testcases import config_state  # noqa
from oc_config_validate.testcases import get  # noqa
from oc_config_validate.testcases import set  # noqa
from oc_config_validate.testcases import setget  # noqa
from oc_config_validate.testcases import static_route  # noqa
from oc_config_validate.testcases import subif_ip  # noqa
from oc_config_validate.testcases import telemetry_once  # noqa
from oc_config_validate.testcases import telemetry_onchange  # noqa
from oc_config_validate.testcases import telemetry_sample  # noqa
